import pygeos

from threedigrid_builder.base import Array
from threedigrid_builder.constants import CalculationType, ContentType
from threedigrid_builder.grid import linear
from threedigrid_builder.grid.cross_section_locations import compute_bottom_level

__all__ = ["Channels"]


class Channel:
    id: int
    code: str
    the_geom: pygeos.Geometry
    dist_calc_points: float
    connection_node_start_id: int
    connection_node_end_id: int
    calculation_type: CalculationType
    display_name: str
    zoom_category: int


class Channels(Array[Channel], linear.BaseLinear):
    content_type = ContentType.TYPE_V2_CHANNEL

    def get_1d2d_properties(self, nodes, node_idx, locations, **kwargs):
        """Compute properties (is_closed, dpumax) of 1D-2D channel flowlines.

        Args:
            nodes (Nodes): All nodes
            node_idx (array of int): indices into nodes for which to compute properties
            locations (CrossSectionLocations): for the bank_levels

        Returns:
            tuple of:
            - is_closed (bool): always False
            - dpumax (array of float): interpolated between CS location bank_levels
        """
        # dpumax is interpolated between cross section location bank levels
        dpumax = compute_bottom_level(
            nodes.content_pk[node_idx],
            nodes.s1d[node_idx],
            locations,
            self,
            "bank_level",
        )

        return False, dpumax



from unittest import mock

import numpy as np
import pygeos
import pytest
from numpy.testing import assert_almost_equal, assert_array_equal

from threedigrid_builder.base import Nodes
from threedigrid_builder.grid import ConnectionNodes, Pipes


@pytest.fixture
def connection_nodes():
    # Used to map connection_node_start/end_id to an index (sequence id)
    return ConnectionNodes(
        id=[21, 25, 33, 42],
        the_geom=pygeos.points([(0, 21), (1, 25), (2, 33), (3, 42)]),
        drain_level=[0.0, 20.0, 30.0, 10.0],
    )


@pytest.fixture
def pipes():
    return Pipes(
        id=[1, 2],
        dist_calc_points=[5.0, np.nan],
        code=["one", "two"],
        connection_node_start_id=[21, 21],
        connection_node_end_id=[25, 42],
        calculation_type=[2, 1],
        invert_level_start_point=[3.0, 5.0],
        invert_level_end_point=[4.0, 6.0],
    )


@pytest.mark.parametrize(
    "pipe_ids,ds,expected",
    [
        ([1], [5.0], [3.5]),
        ([2], [0.5], [5.5]),
        ([1, 2], [5.0, 0.5], [3.5, 5.5]),
        ([2, 1], [0.5, 5.0], [5.5, 3.5]),
        ([1, 1, 2, 2], [5.0, 7.5, 0.5, 0.25], [3.5, 3.75, 5.5, 5.25]),
    ],
)
def test_compute_bottom_level(pipe_ids, ds, pipes, expected):
    # set geometries with lengths 10 and 1 (resp. id 1 and 2)
    # invert levels are [3, 4] for id=1 and [5, 6] for id=2
    pipes.the_geom = pygeos.linestrings([[(0, 0), (0, 10)], [(2, 2), (3, 2)]])

    actual = pipes.compute_bottom_level(pipe_ids, ds)

    assert_almost_equal(actual, expected)


def test_compute_bottom_level_raises_no_geom(pipes):
    with pytest.raises(ValueError, match=".*Call set_geometries first.*"):
        pipes.compute_bottom_level([1], [5.0])


@pytest.mark.parametrize(
    "pipe_ids,ds",
    [
        ([1], [10.1]),
        ([2], [-1e-7]),
    ],
)
def test_compute_bottom_level_raises_out_of_bounds(pipe_ids, ds, pipes):
    pipes.the_geom = pygeos.linestrings([[(0, 0), (0, 10)], [(2, 2), (3, 2)]])
    with pytest.raises(ValueError, match=".*outside of the linear object bounds.*"):
        pipes.compute_bottom_level(pipe_ids, ds)


def test_1d2d_properties(pipes):
    nodes = Nodes(
        id=[0, 2, 5, 7],
        content_pk=[1, 1, 2, 2],
        s1d=[12.0, 13.0, 14.0, 15.0],
    )
    node_idx = [0, 1, 3]

    connection_nodes = mock.Mock()

    with mock.patch.object(pipes, "compute_drain_level") as compute_drain_level:
        compute_drain_level.return_value = np.array([1.8, 2.5, 3.2])
        is_closed, dpumax = pipes.get_1d2d_properties(nodes, node_idx, connection_nodes)

        _, kwargs = compute_drain_level.call_args
        assert_array_equal(kwargs["ids"], [1, 1, 2])  # the content pk
        assert_array_equal(kwargs["s"], [12.0, 13.0, 15.0])  # the s1d
        assert kwargs["connection_nodes"] is connection_nodes

    # pipes are closed
    assert_array_equal(is_closed, True)

    # interpolation between manhole drain levels is further tested elsewhere
    assert_array_equal(dpumax, [1.8, 2.5, 3.2])



import itertools

import numpy as np
import pygeos

from threedigrid_builder.base import Array, Lines, Nodes
from threedigrid_builder.constants import (
    BoundaryType,
    CalculationType,
    ContentType,
    LineType,
    NodeType,
)
from threedigrid_builder.exceptions import SchematisationError

__all__ = ["BoundaryConditions1D", "BoundaryConditions2D"]


class BoundaryCondition1D:
    id: int
    boundary_type: BoundaryType
    connection_node_id: int


class BoundaryConditions1D(Array[BoundaryCondition1D]):
    def apply(self, grid):
        """Apply 1D boundary conditions to a Grid

        Fields on nodes/lines that have a boundary condition be adjusted:
        - nodes.calculation_type: set (overridden) to BOUNDARY_NODE
        - nodes.boundary_id: from BoundaryConditions1D.id
        - nodes.boundary_type: from BoundaryConditions1D.boundary_type
        - nodes.node_type: set to NODE_1D_BOUNDARIES
        - lines.boundary_id: from BoundaryConditions1D.id

        Note:
        - it is assumed that nodes.content_pk is sorted for connection nodes
        - this method should go before 1D-2D connections and node embedding
        """
        # lookup corresponding nodes
        cn_idx = np.where(
            grid.nodes.content_type == ContentType.TYPE_V2_CONNECTION_NODES
        )[0]
        idx = cn_idx[
            np.searchsorted(grid.nodes.content_pk[cn_idx], self.connection_node_id)
        ]
        if not np.all(grid.nodes.content_pk[idx] == self.connection_node_id):
            missing = np.where(grid.nodes.content_pk[idx] != self.connection_node_id)[0]
            raise SchematisationError(
                f"1D boundary conditions {self.id[missing].tolist()} refer to missing "
                f"connection nodes {self.connection_node_id[missing].tolist()}."
            )

        # lookup the lines connecting to these nodes
        ids = grid.nodes.index_to_id(idx)
        line_idx, start_or_end = np.where(np.isin(grid.lines.line, ids))
        # lookup the boundary condition node and other node corresponding to these lines
        bc_node_id = grid.lines.line[line_idx, start_or_end]
        # this trick transforms 0 to 1 and 1 to 0:
        start_or_end_inv = (~(start_or_end.astype(bool))).astype(int)
        other_node_id = grid.lines.line[line_idx, start_or_end_inv]

        if not np.isin(ids, bc_node_id).all():
            missing = ~np.isin(ids, bc_node_id)
            raise SchematisationError(
                f"1D boundary conditions {self.id[missing].tolist()} refer to connection nodes "
                f"that have no attached objects ({self.connection_node_id[missing].tolist()})."
            )
        if len(ids) != len(bc_node_id):
            sorted_cn_id, counts = np.unique(bc_node_id, return_counts=True)
            duplicate_cn_id = sorted_cn_id[counts > 1]
            raise SchematisationError(
                f"Only one attached object is allowed for connection nodes "
                f"{duplicate_cn_id.tolist()} with a 1D boundary condition."
            )
        if np.isin(ids, other_node_id).any():
            both_sides = np.isin(ids, other_node_id)
            raise SchematisationError(
                f"1D boundary conditions {self.id[both_sides].tolist()} are too close to other "
                f"boundary conditions."
            )

        # set node attributes
        grid.nodes.calculation_type[idx] = CalculationType.BOUNDARY_NODE
        grid.nodes.boundary_id[idx] = self.id
        grid.nodes.boundary_type[idx] = self.boundary_type
        grid.nodes.node_type[idx] = NodeType.NODE_1D_BOUNDARIES
        # set line attributes
        grid.lines.is_1d_boundary[line_idx] = 1


class BoundaryCondition2D:
    id: int
    boundary_type: BoundaryType
    the_geom: pygeos.Geometry


class BoundaryConditions2D(Array[BoundaryCondition2D]):
    def get_nodes_and_lines(
        self, nodes, cell_tree, quadtree, node_id_counter, line_id_counter
    ):
        # we pad quad_idx to prevent out of bound errors later
        quad_idx = np.pad(quadtree.quad_idx, 1, mode="constant", constant_values=0)
        cells_done = np.empty((0,), dtype=int)

        boundary_nodes = Nodes(id=[])
        boundary_lines = Lines(id=[])
        for bc_idx in range(len(self)):
            bc_geom = self.the_geom[bc_idx]

            x1, y1, x2, y2 = pygeos.bounds(bc_geom)
            is_horizontal = (x2 - x1) > (y2 - y1)

            node_idx = np.sort(cell_tree.query(bc_geom))
            if node_idx.size == 0:
                # Pragmatic fix; because of reprojection issues, the boundary condition
                # can be right outside the cells.

                # Start by getting the closest point inside the model area (all cells)
                nearest_cell = cell_tree.geometries[cell_tree.nearest(bc_geom)[1][0]]
                shortest_line = pygeos.shortest_line(nearest_cell, bc_geom)
                x1, _, x2, _ = pygeos.bounds(nearest_cell)
                if pygeos.length(shortest_line) < (x2 - x1):
                    cell_coords, _ = pygeos.get_coordinates(shortest_line)

                    # Adjust the x / y coordinate to the x / y coordinate of the point
                    bc_coords = pygeos.get_coordinates(bc_geom)
                    bc_coords[:, int(is_horizontal)] = cell_coords[int(is_horizontal)]
                    bc_geom = pygeos.set_coordinates(bc_geom, bc_coords)

                    # Query again for intersecting cells
                    node_idx = np.sort(cell_tree.query(bc_geom))

            nodk = nodes.nodk[node_idx]
            sz = 2 ** (nodk - 1)
            # nodm and nodn are 1-based indices into quad_idx. So we subtract 1 from it.
            nodm = nodes.nodm[node_idx] - 1
            nodn = nodes.nodn[node_idx] - 1
            # Resized nodm, nodn (called x, y) so that they index into quad_idx.
            # The +1 accounts for the padding of quad_idx.
            x = nodm * sz + 1
            y = nodn * sz + 1
            # just to be sure, can be removed later:
            assert np.all(quad_idx[x, y] == (node_idx + 1))

            if is_horizontal:  # horizontal BC, looking at vertical neigbours
                # A cell has no bottom (upstream) neighbor if:
                # - the nodgrid entry before (under) at the left & right sides is 0
                before = (quad_idx[x, y - 1] == 0) & (quad_idx[x + sz - 1, y - 1] == 0)
                # A cell has no top (downstream) neighbor if:
                # - the nodgrid entry after (above) at the left & right sides is 0
                after = (quad_idx[x, y + sz] == 0) & (quad_idx[x + sz - 1, y + sz] == 0)
            else:  # vertical BC, looking at horizontal neigbours
                # A cell has no left (upstream) neighbor if:
                # - the nodgrid entry before (to the left) at bottom & top sides are 0
                before = (quad_idx[x - 1, y] == 0) & (quad_idx[x - 1, y + sz - 1] == 0)
                # A cell has no right (downstream) neighbor if:
                # - the nodgrid entry after (to the right) at bottom & top sides is 0
                after = (quad_idx[x + sz, y] == 0) & (quad_idx[x + sz, y + sz - 1] == 0)

            # The edge of the boundary is whichever side has the most edges
            n_edges_before = np.count_nonzero(before)
            n_edges_after = np.count_nonzero(after)
            if n_edges_after == 0 and n_edges_before == 0:
                raise SchematisationError(
                    f"2D boundary condition {self.id[bc_idx]} does not touch any edge "
                    f"cell."
                )
            elif n_edges_after == n_edges_before:
                s = "top and bottom" if is_horizontal else "right and left"
                raise SchematisationError(
                    f"2D boundary condition {self.id[bc_idx]} touches cells that have "
                    f"equal numbers of {s} edges."
                )
            is_before = n_edges_before > n_edges_after  # is_upstream

            # Filter the nodes to those that have edges
            if is_horizontal and is_before:
                kcu = LineType.LINE_2D_BOUNDARY_SOUTH
                node_idx = node_idx[before]
                edge_coord = nodes.bounds[node_idx, 1]
                cross_pix_coords_cols = (0, 1, 2, 1)
            elif is_horizontal:  # and not is_before
                kcu = LineType.LINE_2D_BOUNDARY_NORTH
                node_idx = node_idx[after]
                edge_coord = nodes.bounds[node_idx, 3]
                cross_pix_coords_cols = (0, 3, 2, 3)
            elif is_before:  # and not is_horizontal
                kcu = LineType.LINE_2D_BOUNDARY_WEST
                node_idx = node_idx[before]
                edge_coord = nodes.bounds[node_idx, 0]
                cross_pix_coords_cols = (0, 1, 0, 3)
            else:  # not is_horizontal and not is_before
                kcu = LineType.LINE_2D_BOUNDARY_EAST
                node_idx = node_idx[after]
                edge_coord = nodes.bounds[node_idx, 2]
                cross_pix_coords_cols = (2, 1, 2, 3)

            if edge_coord.size > 1 and np.any(edge_coord[1:] != edge_coord[:-1]):
                coords = np.unique(edge_coord)
                axis = "y" if is_horizontal else "x"
                raise SchematisationError(
                    f"2D boundary condition {self.id[bc_idx]} touches cells with "
                    f"different edge coordinates ({axis}={coords.tolist()})."
                )

            # Copy the nodes and change the attributes into a boundary node
            node_idx = node_idx[~np.in1d(node_idx, cells_done)]
            if len(node_idx) == 0:
                continue

            new_nodes = nodes[node_idx]
            new_nodes.id[:] = list(itertools.islice(node_id_counter, len(new_nodes)))
            new_nodes.boundary_id[:] = self.id[bc_idx]
            new_nodes.boundary_type[:] = self.boundary_type[bc_idx]
            new_nodes.node_type[:] = NodeType.NODE_2D_BOUNDARIES

            # Get the cross_pix_coords before resetting the pixel_coords
            cross_pix_coords = np.array(
                [new_nodes.pixel_coords[:, x] for x in cross_pix_coords_cols]
            ).T

            # a boundary cell has no real place on the pixels:
            new_nodes.pixel_coords[:] = -9999

            # the bounds and coordinates are shifted:
            size = new_nodes.bounds[:, 2] - new_nodes.bounds[:, 0]
            axes = (1, 3) if is_horizontal else (0, 2)
            add_or_subtract = -1 if is_before else 1
            new_nodes.bounds[:, axes[0]] += size * add_or_subtract
            new_nodes.bounds[:, axes[1]] += size * add_or_subtract
            new_nodes.coordinates[:, axes[0]] += size * add_or_subtract
            if is_horizontal:
                new_nodes.nodn[:] += add_or_subtract
            else:
                new_nodes.nodm[:] += add_or_subtract

            # Create new lines
            if is_before:
                line = np.array([new_nodes.id, nodes.index_to_id(node_idx)]).T
            else:
                line = np.array([nodes.index_to_id(node_idx), new_nodes.id]).T

            new_lines = Lines(
                id=itertools.islice(line_id_counter, len(new_nodes)),
                kcu=kcu,
                line=line,
                cross_pix_coords=cross_pix_coords,
                lik=new_nodes.nodk,
            )

            cells_done = np.append(cells_done, node_idx)
            boundary_nodes += new_nodes
            boundary_lines += new_lines

        return boundary_nodes, boundary_lines



import numpy as np

from threedigrid_builder.constants import ContentType

from .array import Array
from .lines import Lines
from .nodes import Nodes

__all__ = ["Endpoints"]


class Endpoint:
    id: int
    line_idx: int
    is_start: bool
    node_id: int


class Endpoints(Array[Endpoint]):
    """Each line has 2 endpoints.

    This class makes it easier to reason about all the objects on a node.
    """

    scalars = ("lines",)

    @classmethod
    def for_connection_nodes(
        cls,
        nodes: Nodes,
        lines: Lines,
        line_mask=None,
        node_mask=None,
    ) -> "Endpoints":
        """Return the endpoints for connection nodes

        Optionally filter by the content type of the lines.
        """
        if line_mask is None:
            where = np.arange(len(lines), dtype=int)
        else:
            where = np.where(line_mask)[0]
        n = len(where)
        result = cls(
            lines=lines,
            id=range(n * 2),
            line_idx=np.repeat(where, 2),
            node_id=lines.line[where].ravel(),
            is_start=np.tile([True, False], n),
        )
        is_cn = nodes.content_type == ContentType.TYPE_V2_CONNECTION_NODES
        node_mask = is_cn if node_mask is None else is_cn & node_mask
        return result[np.isin(result.node_id, nodes.id[node_mask])]

    @property
    def is_end(self):
        return ~self.is_start

    @property
    def invert_level(self):
        return np.where(
            self.is_start, self.invert_level_start_point, self.invert_level_end_point
        )

    def __getattr__(self, name):
        return getattr(self.lines, name)[self.line_idx]

    def reduce_per_node(self, reduceat, values) -> "NodeValues":
        """Compute the per-node reduction of 'values'

        The 'reduceat' function is normally a numpy ufunc.reduceat. This function
        is expected to reduce 'values' on index ranges contained in 'indices'. The length
        of 'indices' must equal the length of the returned value.

        For instance:

         - self.node_id = [1, 1, 5, 5, 5]
         - values = [1., 3., 5., 7., 9.]
         - gives: indices = [0, 2]
         - and reduceat is expected to act on indices [0, 2) and [2, <end>)

        Returns node ids and corresponding maximum values.
        """
        if len(values) != len(self):
            raise ValueError("values must have the same length as self")
        if len(values) == 0:
            return NodeValues(id=[])
        diff = np.diff(self.node_id)
        if np.any(diff < 0):
            raise ValueError("endpoints must be ordered by node_id")
        indices = np.concatenate([[0], np.where(diff > 0)[0] + 1])
        result = reduceat(values, indices)
        return NodeValues(id=self.node_id[indices], value=result)

    def nanmin_per_node(self, values) -> "NodeValues":
        return self.reduce_per_node(np.fmin.reduceat, values)

    def first_per_node(self, values) -> "NodeValues":
        return self.reduce_per_node(lambda x, y: x[y], values)


class NodeValue:
    id: int
    value: float


class NodeValues(Array[NodeValue]):
    pass



import numpy as np
import pygeos
import pytest
from numpy.testing import assert_almost_equal, assert_equal

from threedigrid_builder.base import Endpoints, Lines, Nodes
from threedigrid_builder.constants import ContentType

CH = ContentType.TYPE_V2_CHANNEL
CN = ContentType.TYPE_V2_CONNECTION_NODES


@pytest.fixture
def nodes():
    return Nodes(
        id=[1, 2, 3], coordinates=[(1, 1), (2, 2), (3, 3)], content_type=[CN, CN, -9999]
    )


@pytest.fixture
def lines():
    return Lines(
        id=[0, 1, 2],
        line=[(1, 2), (2, 3), (1, 3)],
        line_geometries=[None, pygeos.linestrings([(5, 5), (6, 6)]), None],
        ds1d=[np.nan, 16.0, 2.0],
        content_type=[CN, CH, CH],
    )


def test_for_connection_nodes(nodes: Nodes, lines: Lines):
    endpoints = Endpoints.for_connection_nodes(nodes, lines)
    endpoints.reorder_by("node_id")
    assert isinstance(endpoints, Endpoints)
    assert_equal(endpoints.node_id, [1, 1, 2, 2])
    assert_equal(endpoints.line_idx, [0, 2, 0, 1])
    assert_equal(endpoints.is_start, [1, 1, 0, 1])


def test_for_connection_nodes_with_line_mask(nodes: Nodes, lines: Lines):
    endpoints = Endpoints.for_connection_nodes(
        nodes, lines, line_mask=lines.content_type == CH
    )
    endpoints.reorder_by("node_id")
    assert isinstance(endpoints, Endpoints)
    assert_equal(endpoints.node_id, [1, 2])
    assert_equal(endpoints.line_idx, [2, 1])
    assert_equal(endpoints.is_start, [1, 1])


def test_for_connection_nodes_with_node_mask(nodes: Nodes, lines: Lines):
    endpoints = Endpoints.for_connection_nodes(nodes, lines, node_mask=nodes.id != 2)
    assert isinstance(endpoints, Endpoints)
    assert_equal(endpoints.node_id, [1, 1])
    assert_equal(endpoints.line_idx, [0, 2])
    assert_equal(endpoints.is_start, [1, 1])


def test_getattr(lines: Lines):
    endpoints = Endpoints(lines=lines, id=[0, 1, 2], line_idx=[1, 2, 1])

    assert_equal(endpoints.ds1d, lines.ds1d[[1, 2, 1]])


@pytest.mark.parametrize(
    "ufunc,expected",
    [
        (np.fmax.reduceat, [16.0, 16.0]),
        (np.fmin.reduceat, [2.0, 16.0]),
        (np.add.reduceat, [18.0, np.nan]),
        (lambda x, y: y, [0.0, 2.0]),
    ],
)
def test_reduce_per_node(lines: Lines, ufunc, expected):
    endpoints = Endpoints(
        lines=lines, id=range(4), node_id=[1, 1, 5, 5], line_idx=[1, 2, 1, 0]
    )
    actual = endpoints.reduce_per_node(ufunc, endpoints.ds1d)

    assert_equal(actual.id, [1, 5])
    assert_almost_equal(actual.value, expected)


def test_reduce_per_node_empty(lines: Lines):
    endpoints = Endpoints(lines=lines, id=[])
    actual = endpoints.reduce_per_node(None, [])

    assert_equal(actual.id, [])
    assert_almost_equal(actual.value, [])


def test_nanmin_per_node(lines: Lines):
    endpoints = Endpoints(
        lines=lines, id=range(4), node_id=[1, 1, 5, 5], line_idx=[1, 2, 1, 0]
    )
    actual = endpoints.nanmin_per_node(endpoints.ds1d)

    assert_equal(actual.id, [1, 5])
    assert_almost_equal(actual.value, [2.0, 16.0])


def test_first_per_node(lines: Lines):
    endpoints = Endpoints(
        lines=lines, id=range(4), node_id=[1, 1, 5, 5], line_idx=[1, 2, 0, 1]
    )
    actual = endpoints.first_per_node(endpoints.ds1d)

    assert_equal(actual.id, [1, 5])
    assert_almost_equal(actual.value, [16.0, np.nan])


def test_invert_level(lines: Lines):
    lines.invert_level_start_point[:] = [1, 2, 3]
    lines.invert_level_end_point[:] = [4, 5, 6]

    endpoints = Endpoints(
        lines=lines, id=range(3), line_idx=[0, 2, 1], is_start=[False, True, False]
    )

    assert_almost_equal(endpoints.invert_level, [4, 3, 5])



import pygeos

from threedigrid_builder.base import Array
from threedigrid_builder.constants import CalculationType, ContentType, FrictionType
from threedigrid_builder.grid import linear

__all__ = ["Pipes"]


class Pipe:
    id: int
    code: str
    dist_calc_points: float
    calculation_type: CalculationType
    connection_node_start_id: int
    connection_node_end_id: int
    cross_section_definition_id: int
    invert_level_start_point: float
    invert_level_end_point: float
    sewerage_type: int
    friction_type: FrictionType
    friction_value: float
    the_geom: pygeos.Geometry
    zoom_category: int
    display_name: str
    # profile_num
    # original_length
    material: int


class Pipes(Array[Pipe], linear.BaseLinear):
    content_type = ContentType.TYPE_V2_PIPE

    def get_1d2d_properties(self, nodes, node_idx, connection_nodes, **kwargs):
        """Compute properties (is_closed, dpumax) of 1D-2D pipe flowlines.

        Args:
            nodes (Nodes): All nodes
            node_idx (array of int): indices into nodes for which to compute properties
            connection_nodes (ConnectionNodes): for the drain_level

        Returns:
            tuple of:
            - is_closed (bool): always True
            - dpumax (array of float): interpolated between CN drain_level
        """
        # dpumax is interpolated between drain levels of adjacent manholes (conn nodes)
        dpumax = self.compute_drain_level(
            ids=nodes.content_pk[node_idx],
            s=nodes.s1d[node_idx],
            connection_nodes=connection_nodes,
        )

        return True, dpumax



import numpy as np
import pygeos
import pytest
from numpy.testing import assert_array_equal

from threedigrid_builder.base import Nodes
from threedigrid_builder.constants import ContentType
from threedigrid_builder.grid import Channels, CrossSectionLocations


@pytest.fixture
def channels():
    return Channels(
        the_geom=pygeos.linestrings([[(0, 0), (6, 0), (6, 6)]]),
        dist_calc_points=np.array([5.0]),
        id=np.array([1]),
        code=np.array(["one"]),
        connection_node_start_id=np.array([21]),
        connection_node_end_id=np.array([42]),
        calculation_type=np.array([2]),
    )


def test_1d2d_properties(channels):
    nodes = Nodes(
        id=[0, 2, 5],
        s1d=[3, np.nan, 6],
        content_type=[
            ContentType.TYPE_V2_CHANNEL,
            ContentType.TYPE_V2_CONNECTION_NODES,
            ContentType.TYPE_V2_CHANNEL,
        ],
        content_pk=[1, 3, 1],
    )
    locations = CrossSectionLocations(
        id=[2, 5],
        the_geom=pygeos.points([(0, 0), [6, 6]]),
        bank_level=[1.0, 13.0],
        channel_id=[1, 1],
    )
    node_idx = [0, 2]

    is_closed, dpumax = channels.get_1d2d_properties(nodes, node_idx, locations)

    # channels are no sewerage
    assert_array_equal(is_closed, False)

    # bank levels are interpolated
    assert_array_equal(dpumax, [4.0, 7.0])
